@@ -260,6 +260,10 @@ struct bch_sb *bch2_format(struct bch_opt_strs	fs_opt_strs,
 	sb.sb->nr_devices	= devs.nr;
 	SET_BCH_SB_VERSION_INCOMPAT_ALLOWED(sb.sb, opts.version);
 
+	/* These are no longer options, only provided for compatibility with old verions */
+	SET_BCH_SB_META_REPLICAS_REQ(sb.sb, 1);
+	SET_BCH_SB_DATA_REPLICAS_REQ(sb.sb, 1);
+
 	if (opts.version > bcachefs_metadata_version_disk_accounting_big_endian)
 		sb.sb->features[0] |= cpu_to_le64(BCH_SB_FEATURES_ALL);
 
@@ -736,7 +740,7 @@ const struct bch_option *bch2_cmdline_opt_parse(int argc, char *argv[],
 		return NULL;
 
 	char *optstr = strdup(argv[optind] + 2);
-	optarg = argv[optind + 1];
+	optarg = NULL;
 
 	char *eq = strchr(optstr, '=');
 	if (eq) {
@@ -744,10 +748,6 @@ const struct bch_option *bch2_cmdline_opt_parse(int argc, char *argv[],
 		optarg = eq + 1;
 	}
 
-	if (!optarg)
-		optarg = "1";
-
-
 	int optid = bch2_opt_lookup(optstr);
 	if (optid < 0)
 		goto noopt;
@@ -758,11 +758,11 @@ const struct bch_option *bch2_cmdline_opt_parse(int argc, char *argv[],
 
 	optind++;
 
-	if (opt->type != BCH_OPT_BOOL) {
-		if (optarg == argv[optind])
-			optind++;
-	} else {
-		optarg = NULL;
+	if (!optarg) {
+		if (opt->type != BCH_OPT_BOOL)
+			optarg = argv[optind++];
+		else
+			optarg = "1";
 	}
 
 	return opt;
