@@ -248,10 +248,10 @@
 
 #include "data/compress_types.h"
 #include "data/copygc_types.h"
-#include "data/ec_types.h"
+#include "data/ec/types.h"
 #include "data/keylist_types.h"
 #include "data/nocow_locking_types.h"
-#include "data/reconcile_types.h"
+#include "data/reconcile/types.h"
 
 #include "debug/async_objs_types.h"
 #include "debug/trace.h"
@@ -550,7 +550,7 @@ enum bch_dev_read_ref {
 
 #define BCH_DEV_WRITE_REFS()				\
 	x(journal_write)				\
-	x(journal_do_discards)				\
+	x(journal_discard)				\
 	x(dev_do_discards)				\
 	x(discard_one_bucket_fast)			\
 	x(do_invalidates)				\
@@ -791,7 +791,8 @@ struct bch_fs {
 	struct bch_devs_mask	devs_rotational;
 
 	struct bch_opts		opts;
-	atomic_t		opt_change_cookie;
+	struct mutex		opt_change_lock;
+	u32			opt_change_cookie;
 	struct bch_opts_mask	mount_opts;
 
 	struct bch_sb_cpu	sb;
