@@ -164,14 +164,12 @@ __aligned(4)
 #define KEY_SNAPSHOT_MAX		((__u32)~0U)
 #define KEY_SIZE_MAX			((__u32)~0U)
 
-static inline struct bpos SPOS(__u64 inode, __u64 offset, __u32 snapshot)
-{
-	return (struct bpos) {
-		.inode		= inode,
-		.offset		= offset,
-		.snapshot	= snapshot,
-	};
-}
+#define SPOS(_inode, _offset, _snapshot)		\
+	((struct bpos) {				\
+		.inode		= _inode,		\
+		.offset		= _offset,		\
+		.snapshot	= _snapshot,		\
+	})
 
 #define POS_MIN				SPOS(0, 0, 0)
 #define POS_MAX				SPOS(KEY_INODE_MAX, KEY_OFFSET_MAX, 0)
@@ -453,9 +451,10 @@ struct bch_error {
 };
 
 #define KEY_TYPE_ERRORS()			\
-	x(unknown,		0)		\
-	x(device_removed,	1)		\
-	x(double_allocation,	2)
+	x(unknown,			0)	\
+	x(device_removed,		1)	\
+	x(double_allocation,		2)	\
+	x(no_valid_pointers_repair,	3)
 
 enum bch_key_type_errors {
 #define x(n, t)	KEY_TYPE_ERROR_##n = t,
@@ -646,7 +645,7 @@ enum btree_id {
 #include "alloc/lru_format.h"
 #include "alloc/replicas_format.h"
 #include "alloc/format.h"
-#include "data/ec_format.h"
+#include "data/ec/format.h"
 #include "data/extents_format.h"
 #include "data/extents_sb_format.h"
 #include "data/reflink_format.h"
@@ -847,7 +846,8 @@ struct bch_sb_field_ext {
 	x(sb_field_extent_type_u64s,	BCH_VERSION(1, 32))		\
 	x(reconcile,			BCH_VERSION(1, 33))		\
 	x(extented_key_type_error,	BCH_VERSION(1, 34))		\
-	x(bucket_stripe_index,		BCH_VERSION(1, 35))
+	x(bucket_stripe_index,		BCH_VERSION(1, 35))		\
+	x(no_sb_user_data_replicas,	BCH_VERSION(1, 36))
 
 enum bcachefs_metadata_version {
 	bcachefs_metadata_version_min = 9,
