@@ -26,28 +26,34 @@ enum bch_data_update_types {
 
 struct data_update_opts {
 	enum bch_data_update_types	type;
-	u8				ptrs_rewrite;
 	u8				ptrs_io_error;
 	u8				ptrs_kill;
 	u8				ptrs_kill_ec;
 	u8				extra_replicas;
 	u16				target;
-	int				read_dev;
-	bool				checksum_paranoia;
+	bool				no_devs_have:1;
+	bool				checksum_paranoia:1;
 
+	unsigned			read_dev;
+	enum bch_read_flags		read_flags;
 	enum bch_write_flags		write_flags;
 	enum bch_trans_commit_flags	commit_flags;
 };
 
 struct data_update {
+	struct rcu_head		rcu;
 	/* extent being updated: */
 	enum btree_id		btree_id;
 	struct bkey_buf		k;
 	struct data_update_opts	opts;
 
+	bool			on_hashtable;
 	bool			read_done;
 	u8			ptrs_held;
 
+	struct rhlist_head	hash;
+	struct bbpos		pos;
+
 	/* associated with @ctxt */
 	struct list_head	read_list;
 	struct list_head	io_list;
@@ -61,16 +67,12 @@ struct data_update {
 };
 
 struct promote_op {
-	struct rcu_head		rcu;
 	u64			start_time;
 #ifdef CONFIG_BCACHEFS_ASYNC_OBJECT_LISTS
 	unsigned		list_idx;
 #endif
 	int			cpu; /* for promote_limit */
 
-	struct rhash_head	hash;
-	struct bpos		pos;
-
 	struct work_struct	work;
 	struct data_update	write;
 	struct bio_vec		bi_inline_vecs[]; /* must be last */
@@ -80,17 +82,15 @@ void bch2_data_update_opts_to_text(struct printbuf *, struct bch_fs *,
 				   struct bch_inode_opts *, struct data_update_opts *);
 void bch2_data_update_to_text(struct printbuf *, struct data_update *);
 void bch2_data_update_inflight_to_text(struct printbuf *, struct data_update *);
+bool bch2_data_update_in_flight(struct bch_fs *, struct bbpos *);
 
 int bch2_data_update_index_update(struct bch_write_op *);
 
 void bch2_data_update_read_done(struct data_update *);
 
-struct bch_devs_list bch2_data_update_devs_keeping(struct bch_fs *,
-						   struct data_update_opts *,
-						   struct bkey_s_c);
-int bch2_can_do_write(struct bch_fs *, struct bch_inode_opts *,
-		      struct data_update_opts *,
-		      struct bkey_s_c, struct bch_devs_list *);
+int bch2_can_do_data_update(struct btree_trans *, struct bch_inode_opts *,
+			    struct data_update_opts *, struct bkey_s_c,
+			    struct printbuf *);
 
 void bch2_data_update_exit(struct data_update *, int);
 int bch2_data_update_init(struct btree_trans *, struct btree_iter *,
@@ -100,4 +100,7 @@ int bch2_data_update_init(struct btree_trans *, struct btree_iter *,
 			  struct bch_inode_opts *, struct data_update_opts,
 			  enum btree_id, struct bkey_s_c);
 
+void bch2_fs_data_update_exit(struct bch_fs *);
+int bch2_fs_data_update_init(struct bch_fs *);
+
 #endif /* _BCACHEFS_DATA_UPDATE_H */
