@@ -15,12 +15,12 @@
 
 #include "data/checksum.h"
 #include "data/compress.h"
-#include "data/ec.h"
+#include "data/ec/create.h"
 #include "data/extent_update.h"
 #include "data/keylist.h"
 #include "data/move.h"
 #include "data/nocow_locking.h"
-#include "data/rebalance.h"
+#include "data/reconcile/trigger.h"
 #include "data/write.h"
 
 #include "debug/async_objs.h"
@@ -113,50 +113,41 @@ void bch2_latency_acct(struct bch_dev *ca, u64 submit_time, int rw)
 
 void bch2_bio_free_pages_pool(struct bch_fs *c, struct bio *bio)
 {
-	struct bvec_iter_all iter;
-	struct bio_vec *bv;
+	for (struct bio_vec *bv = bio->bi_io_vec;
+	     bv < bio->bi_io_vec + bio->bi_vcnt;
+	     bv++) {
+		void *p = bvec_virt(bv);
 
-	bio_for_each_segment_all(bv, bio, iter)
-		if (bv->bv_page != ZERO_PAGE(0))
-			mempool_free(bv->bv_page, &c->bio_bounce_pages);
+		if (bv->bv_len == BIO_BOUNCE_BUF_POOL_LEN)
+			mempool_free(p, &c->bio_bounce_bufs);
+		else
+			free_pages((unsigned long) p, get_order(bv->bv_len));
+	}
 	bio->bi_vcnt = 0;
 }
 
-static struct page *__bio_alloc_page_pool(struct bch_fs *c, bool *using_mempool)
+static void __bch2_bio_alloc_pages_pool(struct bch_fs *c, struct bio *bio,
+					unsigned bs, size_t size)
 {
-	struct page *page;
+	mutex_lock(&c->bio_bounce_pages_lock);
 
-	if (likely(!*using_mempool)) {
-		page = alloc_page(GFP_NOFS);
-		if (unlikely(!page)) {
-			mutex_lock(&c->bio_bounce_pages_lock);
-			*using_mempool = true;
-			goto pool_alloc;
+	while (bio->bi_iter.bi_size < size)
+		bio_add_virt_nofail(bio,
+				    mempool_alloc(&c->bio_bounce_bufs, GFP_NOFS),
+				    BIO_BOUNCE_BUF_POOL_LEN);
 
-		}
-	} else {
-pool_alloc:
-		page = mempool_alloc(&c->bio_bounce_pages, GFP_NOFS);
-	}
+	bio->bi_iter.bi_size = min(bio->bi_iter.bi_size, size);
 
-	return page;
+	mutex_unlock(&c->bio_bounce_pages_lock);
 }
 
 void bch2_bio_alloc_pages_pool(struct bch_fs *c, struct bio *bio,
-			       size_t size)
+			       unsigned bs, size_t size)
 {
-	bool using_mempool = false;
+	bch2_bio_alloc_pages(bio, c->opts.block_size, size, GFP_NOFS);
 
-	while (size) {
-		struct page *page = __bio_alloc_page_pool(c, &using_mempool);
-		unsigned len = min_t(size_t, PAGE_SIZE, size);
-
-		BUG_ON(!bio_add_page(bio, page, len, 0));
-		size -= len;
-	}
-
-	if (using_mempool)
-		mutex_unlock(&c->bio_bounce_pages_lock);
+	if (bio->bi_iter.bi_size < size)
+		__bch2_bio_alloc_pages_pool(c, bio, bs, size);
 }
 
 /* Extent update path: */
@@ -266,14 +257,12 @@ static inline int bch2_extent_update_i_size_sectors(struct btree_trans *trans,
 		s64 bi_sectors = le64_to_cpu(inode->v.bi_sectors);
 		if (unlikely(bi_sectors + i_sectors_delta < 0)) {
 			struct bch_fs *c = trans->c;
-			CLASS(printbuf, buf)();
-			bch2_log_msg_start(c, &buf);
-			prt_printf(&buf, "inode %llu i_sectors underflow: %lli + %lli < 0",
+
+			CLASS(bch_log_msg, msg)(c);
+			prt_printf(&msg.m, "inode %llu i_sectors underflow: %lli + %lli < 0",
 				   extent_iter->pos.inode, bi_sectors, i_sectors_delta);
 
-			bool print = bch2_count_fsck_err(c, inode_i_sectors_underflow, &buf);
-			if (print)
-				bch2_print_str(c, KERN_ERR, buf.buf);
+			msg.m.suppress = !bch2_count_fsck_err(c, inode_i_sectors_underflow, &msg.m);
 
 			if (i_sectors_delta < 0)
 				i_sectors_delta = -bi_sectors;
@@ -355,8 +344,8 @@ int bch2_extent_update(struct btree_trans *trans,
 
 	bch2_inode_opts_get_inode(c, &inode, &opts);
 
-	try(bch2_bkey_set_needs_rebalance(c, &opts, k,
-					  SET_NEEDS_REBALANCE_foreground,
+	try(bch2_bkey_set_needs_reconcile(trans, NULL, &opts, k,
+					  SET_NEEDS_RECONCILE_foreground,
 					  change_cookie));
 	try(bch2_trans_update(trans, iter, k, 0));
 	try(bch2_trans_commit(trans, disk_res, NULL,
@@ -390,6 +379,13 @@ static int bch2_write_index_default(struct bch_write_op *op)
 		bch2_trans_begin(trans);
 
 		k = bch2_keylist_front(keys);
+
+		/*
+		 * If we did a degraded write, bch2_bkey_set_needs_reconcile() will add
+		 * pointers to BCH_SB_MEMBER_INVALID so the extent is accounted as
+		 * degraded
+		 */
+		bch2_bkey_buf_realloc(&sk, k->k.u64s + 1 + BCH_REPLICAS_MAX);
 		bch2_bkey_buf_copy(&sk, k);
 
 		int ret = bch2_subvolume_get_snapshot(trans, inum.subvol, &sk.k->k.p.snapshot);
@@ -424,35 +420,31 @@ static int bch2_write_index_default(struct bch_write_op *op)
 
 /* Writes */
 
-void bch2_write_op_error(struct bch_write_op *op, u64 offset, const char *fmt, ...)
+static void bch2_log_write_error_start(struct printbuf *out, bool full,
+				       struct bch_write_op *op, u64 offset)
 {
-	CLASS(btree_trans, trans)(op->c);
-
-	CLASS(printbuf, buf)();
-	bch2_log_msg_start(op->c, &buf);
+	prt_printf(out, "error writing data at ");
 
 	struct bpos pos = op->pos;
 	pos.offset = offset;
 
-	bch2_inum_offset_err_msg_trans(trans, &buf, op->subvol, pos);
+	CLASS(btree_trans, trans)(op->c);
+	bch2_inum_offset_err_msg_trans(trans, out, op->subvol, pos);
+	prt_newline(out);
 
-	prt_str(&buf, "write error: ");
+	bch2_write_op_to_text(out, op);
+}
+
+void bch2_write_op_error(struct bch_write_op *op, bool full, u64 offset, const char *fmt, ...)
+{
+	CLASS(bch_log_msg_ratelimited, msg)(op->c);
+
+	bch2_log_write_error_start(&msg.m, full, op, offset);
 
 	va_list args;
 	va_start(args, fmt);
-	prt_vprintf(&buf, fmt, args);
+	prt_vprintf(&msg.m, fmt, args);
 	va_end(args);
-	prt_newline(&buf);
-
-	if (op->flags & BCH_WRITE_move) {
-		struct data_update *u = container_of(op, struct data_update, op);
-
-		prt_printf(&buf, "from internal move ");
-		bch2_bkey_val_to_text(&buf, op->c, bkey_i_to_s_c(u->k.k));
-		prt_newline(&buf);
-	}
-
-	bch2_print_str_ratelimited(op->c, KERN_ERR, buf.buf);
 }
 
 void bch2_submit_wbio_replicas(struct bch_write_bio *wbio, struct bch_fs *c,
@@ -471,9 +463,15 @@ void bch2_submit_wbio_replicas(struct bch_write_bio *wbio, struct bch_fs *c,
 
 	const struct bch_extent_ptr *last = NULL;
 	bkey_for_each_ptr(ptrs, ptr)
-		last = ptr;
+		if (ptr->dev != BCH_SB_MEMBER_INVALID)
+			last = ptr;
+
+	BUG_ON(!last);
 
 	bkey_for_each_ptr(ptrs, ptr) {
+		if (ptr->dev == BCH_SB_MEMBER_INVALID)
+			continue;
+
 		/*
 		 * XXX: btree writes should be using io_ref[WRITE], but we
 		 * aren't retrying failed btree writes yet (due to device
@@ -561,8 +559,8 @@ static noinline int bch2_write_drop_io_error_ptrs(struct bch_write_op *op)
 		n = bkey_next(src);
 
 		if (bkey_extent_is_direct_data(&src->k)) {
-			bch2_bkey_drop_ptrs(bkey_i_to_s(src), p, entry,
-					    test_bit(p.ptr.dev, op->failed.d));
+			bch2_bkey_drop_ptrs_noerror(bkey_i_to_s(src), p, entry,
+				bch2_dev_io_failures(&op->wbio.failed, p.ptr.dev));
 
 			if (!bch2_bkey_nr_dirty_ptrs(c, bkey_i_to_s_c(src)))
 				return bch_err_throw(c, data_write_io);
@@ -585,11 +583,31 @@ static void __bch2_write_index(struct bch_write_op *op)
 {
 	struct bch_fs *c = op->c;
 	struct keylist *keys = &op->insert_keys;
-	unsigned dev;
 	int ret = 0;
 
 	if (unlikely(op->io_error)) {
 		ret = bch2_write_drop_io_error_ptrs(op);
+
+		CLASS(bch_log_msg, msg)(c);
+
+		/* Separate ratelimit_states for hard and soft errors */
+		msg.m.suppress = !ret
+			? bch2_ratelimit(c)
+			: bch2_ratelimit(c);
+
+		struct bkey_i *k = bch2_keylist_front(&op->insert_keys);
+		bch2_log_write_error_start(&msg.m, false, op, bkey_start_offset(&k->k));
+		bch2_io_failures_to_text(&msg.m, c, &op->wbio.failed);
+
+		if (!ret) {
+			prt_printf(&msg.m, "wrote degraded to ");
+			struct bch_devs_list d = bch2_bkey_devs(c, bkey_i_to_s_c(k));
+			bch2_devs_list_to_text(&msg.m, c, &d);
+			prt_newline(&msg.m);
+		} else {
+			prt_printf(&msg.m, "error %s\n", bch2_err_str(ret));
+		}
+
 		if (ret)
 			goto err;
 	}
@@ -609,7 +627,7 @@ static void __bch2_write_index(struct bch_write_op *op)
 		if (unlikely(ret && !bch2_err_matches(ret, EROFS))) {
 			struct bkey_i *insert = bch2_keylist_front(&op->insert_keys);
 
-			bch2_write_op_error(op, bkey_start_offset(&insert->k),
+			bch2_write_op_error(op, false, bkey_start_offset(&insert->k),
 					    "btree update error: %s", bch2_err_str(ret));
 		}
 
@@ -618,8 +636,9 @@ static void __bch2_write_index(struct bch_write_op *op)
 	}
 out:
 	/* If some a bucket wasn't written, we can't erasure code it: */
-	for_each_set_bit(dev, op->failed.d, BCH_SB_MEMBERS_MAX)
-		bch2_open_bucket_write_error(c, &op->open_buckets, dev, -BCH_ERR_data_write_io);
+	darray_for_each(op->wbio.failed, i)
+		bch2_open_bucket_write_error(c, &op->open_buckets, i->dev,
+					     i->errcode ?: -BCH_ERR_data_write_io);
 
 	bch2_open_buckets_put(c, &op->open_buckets);
 	return;
@@ -734,19 +753,9 @@ static void bch2_write_endio(struct bio *bio)
 				   wbio->submit_time, !bio->bi_status);
 
 	if (unlikely(bio->bi_status)) {
-		if (ca)
-			bch_err_inum_offset_ratelimited(ca,
-					    op->pos.inode,
-					    wbio->inode_offset << 9,
-					    "data write error: %s",
-					    bch2_blk_status_to_str(bio->bi_status));
-		else
-			bch_err_inum_offset_ratelimited(c,
-					    op->pos.inode,
-					    wbio->inode_offset << 9,
-					    "data write error: %s",
-					    bch2_blk_status_to_str(bio->bi_status));
-		set_bit(wbio->dev, op->failed.d);
+		guard(spinlock_irqsave)(&c->write_error_lock);
+		bch2_dev_io_failures_mut(&op->wbio.failed, wbio->dev)->errcode =
+			__bch2_err_throw(c, -blk_status_to_bch_err(bio->bi_status));
 		op->io_error = true;
 	}
 
@@ -838,23 +847,22 @@ static struct bio *bch2_write_bio_alloc(struct bch_fs *c,
 		return bio;
 	}
 
-	wbio->bounce		= true;
+	wbio->bounce = true;
+
 
 	/*
 	 * We can't use mempool for more than c->sb.encoded_extent_max
 	 * worth of pages, but we'd like to allocate more if we can:
 	 */
-	bch2_bio_alloc_pages_pool(c, bio,
-				  min_t(unsigned, output_available,
-					c->opts.encoded_extent_max));
+	bch2_bio_alloc_pages(bio,
+			     c->opts.block_size,
+			     output_available,
+			     GFP_NOFS);
 
-	if (bio->bi_iter.bi_size < output_available)
-		*page_alloc_failed =
-			bch2_bio_alloc_pages(bio,
-					     c->opts.block_size,
-					     output_available -
-					     bio->bi_iter.bi_size,
-					     GFP_NOFS) != 0;
+	unsigned required = min(output_available, c->opts.encoded_extent_max);
+
+	if (unlikely(bio->bi_iter.bi_size < required))
+		__bch2_bio_alloc_pages_pool(c, bio, c->opts.block_size, required);
 
 	return bio;
 }
@@ -958,7 +966,7 @@ static noinline int bch2_write_prep_encoded_data(struct bch_write_op *op, struct
 
 	return 0;
 csum_err:
-	bch2_write_op_error(op, op->pos.offset,
+	bch2_write_op_error(op, false, op->pos.offset,
 		"error verifying existing checksum while moving existing data (memory corruption?)\n"
 		"  expected %0llx:%0llx got %0llx:%0llx type %s",
 		op->crc.csum.hi,
@@ -1051,7 +1059,8 @@ static int bch2_write_extent(struct bch_write_op *op, struct write_point *wp,
 			? BCH_COMPRESSION_TYPE_incompressible
 			: op->compression_opt
 			? bch2_bio_compress(c, dst, &dst_len, src, &src_len,
-					    op->compression_opt)
+					    op->compression_opt,
+					    op->pos, !(op->flags & BCH_WRITE_pages_stable))
 			: 0;
 		if (!crc_is_compressed(crc)) {
 			dst_len = min(dst->bi_iter.bi_size, src->bi_iter.bi_size);
@@ -1207,7 +1216,9 @@ static bool bch2_extent_is_writeable(struct bch_write_op *op,
 		if (crc_is_encoded(p.crc) || p.has_ec)
 			return false;
 
-		replicas += bch2_extent_ptr_durability(c, &p);
+		replicas += !p.ptr.cached
+			? bch2_dev_durability(c, p.ptr.dev)
+			: 0;
 	}
 
 	return replicas >= op->opts.data_replicas;
@@ -1227,8 +1238,15 @@ static int bch2_nocow_write_convert_one_unwritten(struct btree_trans *trans,
 		return 0;
 	}
 
+	/*
+	 * If we did a degraded write, bch2_bkey_set_needs_reconcile() will add
+	 * pointers to BCH_SB_MEMBER_INVALID so the extent is accounted as
+	 * degraded
+	 */
 	struct bkey_i *new = errptr_try(bch2_trans_kmalloc_nomemzero(trans,
-				bkey_bytes(k.k) + sizeof(struct bch_extent_rebalance)));
+				bkey_bytes(k.k) +
+				sizeof(struct bch_extent_reconcile) +
+				sizeof(struct bch_extent_ptr) * BCH_REPLICAS_MAX));
 
 	bkey_reassemble(new, k);
 	bch2_cut_front(c, bkey_start_pos(&orig->k), new);
@@ -1247,7 +1265,7 @@ static int bch2_nocow_write_convert_one_unwritten(struct btree_trans *trans,
 	 */
 
 	/*
-	 * For transactional consistency, set_needs_rebalance() has to be called
+	 * For transactional consistency, set_needs_reconcile() has to be called
 	 * with the io_opts from the btree in the same transaction:
 	 */
 	struct bch_inode_unpacked inode;
@@ -1256,8 +1274,8 @@ static int bch2_nocow_write_convert_one_unwritten(struct btree_trans *trans,
 	return  bch2_extent_update_i_size_sectors(trans, iter,
 					min(new->k.p.offset << 9, new_i_size), 0, &inode) ?:
 		(bch2_inode_opts_get_inode(c, &inode, &opts),
-		 bch2_bkey_set_needs_rebalance(c, &opts, new,
-					       SET_NEEDS_REBALANCE_foreground,
+		 bch2_bkey_set_needs_reconcile(trans, NULL, &opts, new,
+					       SET_NEEDS_RECONCILE_foreground,
 					       op->opts.change_cookie)) ?:
 		bch2_trans_update(trans, iter, new,
 				  BTREE_UPDATE_internal_snapshot_node);
@@ -1266,25 +1284,27 @@ static int bch2_nocow_write_convert_one_unwritten(struct btree_trans *trans,
 static void bch2_nocow_write_convert_unwritten(struct bch_write_op *op)
 {
 	struct bch_fs *c = op->c;
-	struct btree_trans *trans = bch2_trans_get(c);
 	int ret = 0;
 
-	for_each_keylist_key(&op->insert_keys, orig) {
-		ret = for_each_btree_key_max_commit(trans, iter, BTREE_ID_extents,
-				     bkey_start_pos(&orig->k), orig->k.p,
-				     BTREE_ITER_intent, k,
-				     NULL, NULL, BCH_TRANS_COMMIT_no_enospc, ({
-			bch2_nocow_write_convert_one_unwritten(trans, &iter, op, orig, k, op->new_i_size);
-		}));
-		if (ret)
-			break;
-	}
+	{
+		CLASS(btree_trans, trans)(c);
 
-	bch2_trans_put(trans);
+		for_each_keylist_key(&op->insert_keys, orig) {
+			ret = for_each_btree_key_max_commit(trans, iter, BTREE_ID_extents,
+					     bkey_start_pos(&orig->k), orig->k.p,
+					     BTREE_ITER_intent, k,
+					     &op->res, NULL,
+					     BCH_TRANS_COMMIT_no_enospc, ({
+				bch2_nocow_write_convert_one_unwritten(trans, &iter, op, orig, k, op->new_i_size);
+			}));
+			if (ret)
+				break;
+		}
+	}
 
 	if (ret && !bch2_err_matches(ret, EROFS)) {
 		struct bkey_i *insert = bch2_keylist_front(&op->insert_keys);
-		bch2_write_op_error(op, bkey_start_offset(&insert->k),
+		bch2_write_op_error(op, false, bkey_start_offset(&insert->k),
 				    "btree update error: %s", bch2_err_str(ret));
 	}
 
@@ -1328,7 +1348,24 @@ static bool bkey_get_dev_iorefs(struct bch_fs *c, struct bkey_ptrs_c ptrs)
 	return true;
 }
 
-static void bch2_nocow_write(struct bch_write_op *op)
+static int bch2_inode_get_i_size(struct btree_trans *trans, struct bpos inode_pos, u64 *i_size)
+{
+	CLASS(btree_iter, iter)(trans, BTREE_ID_inodes, inode_pos, BTREE_ITER_cached);
+	struct bkey_s_c k = bkey_try(bch2_btree_iter_peek_slot(&iter));
+
+	if (likely(k.k->type == KEY_TYPE_inode_v3)) {
+		*i_size = le64_to_cpu(bkey_s_c_to_inode_v3(k).v->bi_size);
+	} else {
+		struct bch_inode_unpacked inode_u;
+		bch2_inode_unpack(k, &inode_u);
+		*i_size = inode_u.bi_size;
+	}
+
+	return 0;
+}
+
+/* returns false if fallaback to cow write path required */
+static bool bch2_nocow_write(struct bch_write_op *op)
 {
 	struct bch_fs *c = op->c;
 	struct btree_trans *trans;
@@ -1340,7 +1377,9 @@ static void bch2_nocow_write(struct bch_write_op *op)
 	int stale, ret;
 
 	if (op->flags & BCH_WRITE_move)
-		return;
+		return false;
+
+	op->flags &= ~BCH_WRITE_convert_unwritten;
 
 	trans = bch2_trans_get(c);
 retry:
@@ -1350,6 +1389,14 @@ retry:
 	if (unlikely(ret))
 		goto err;
 
+	u64 i_size;
+	ret = bch2_inode_get_i_size(trans, SPOS(0, op->pos.inode, snapshot), &i_size);
+	if (unlikely(ret))
+		goto err;
+
+	if (op->new_i_size > i_size)
+		op->flags |= BCH_WRITE_convert_unwritten;
+
 	bch2_trans_iter_init(trans, &iter, BTREE_ID_extents,
 			     SPOS(op->pos.inode, op->pos.offset, snapshot),
 			     BTREE_ITER_slots);
@@ -1360,6 +1407,7 @@ retry:
 		if (ret)
 			break;
 
+		bch2_btree_iter_set_pos(&iter, SPOS(op->pos.inode, op->pos.offset, snapshot));
 		k = bch2_btree_iter_peek_slot(&iter);
 		ret = bkey_err(k);
 		if (ret)
@@ -1410,8 +1458,7 @@ retry:
 		}
 
 		bch2_cut_front(c, op->pos, op->insert_keys.top);
-		if (op->flags & BCH_WRITE_convert_unwritten)
-			bch2_cut_back(POS(op->pos.inode, op->pos.offset + bio_sectors(bio)), op->insert_keys.top);
+		bch2_cut_back(POS(op->pos.inode, op->pos.offset + bio_sectors(bio)), op->insert_keys.top);
 
 		bio = &op->wbio.bio;
 		if (k.k->p.offset < op->pos.offset + bio_sectors(bio)) {
@@ -1434,10 +1481,10 @@ retry:
 		bch2_submit_wbio_replicas(to_wbio(bio), c, BCH_DATA_user,
 					  op->insert_keys.top, true);
 
-		bch2_keylist_push(&op->insert_keys);
+		if (op->flags & BCH_WRITE_convert_unwritten)
+			bch2_keylist_push(&op->insert_keys);
 		if (op->flags & BCH_WRITE_submitted)
 			break;
-		bch2_btree_iter_advance(&iter);
 	}
 out:
 	bch2_trans_iter_exit(&iter);
@@ -1447,14 +1494,15 @@ err:
 
 	bch2_trans_put(trans);
 	if (ret) {
-		bch2_write_op_error(op, op->pos.offset,
+		bch2_write_op_error(op, false, op->pos.offset,
 				    "%s(): btree lookup error: %s", __func__, bch2_err_str(ret));
 		op->error = ret;
 		op->flags |= BCH_WRITE_submitted;
 	}
 
-	/* fallback to cow write path? */
-	if (!(op->flags & BCH_WRITE_submitted)) {
+	bool submitted = op->flags & BCH_WRITE_submitted;
+	if (!submitted) {
+		/* fallback to cow write path */
 		closure_sync(&op->cl);
 		__bch2_nocow_write_done(op);
 		op->insert_keys.top = op->insert_keys.keys;
@@ -1469,7 +1517,7 @@ err:
 		 */
 		continue_at(&op->cl, bch2_nocow_write_done, index_update_wq(op));
 	}
-	return;
+	return submitted;
 err_bucket_stale:
 	{
 		CLASS(printbuf, buf)();
@@ -1498,18 +1546,27 @@ static void __bch2_write(struct bch_write_op *op)
 	struct bch_fs *c = op->c;
 	struct write_point *wp = NULL;
 	struct bio *bio = NULL;
-	unsigned nofs_flags;
 	int ret;
 
-	nofs_flags = memalloc_nofs_save();
+	/*
+	 * Sync or no?
+	 *
+	 * If we're running asynchronously, wne may still want to block
+	 * synchronously here if we weren't able to submit all of the IO at
+	 * once, as that signals backpressure to the caller.
+	 */
+	bool wait_on_allocator_sync = (op->flags & BCH_WRITE_sync) ||
+		(!(op->flags & BCH_WRITE_submitted) &&
+		 !(op->flags & BCH_WRITE_in_worker));
 
-	if (unlikely(op->opts.nocow && c->opts.nocow_enabled)) {
-		bch2_nocow_write(op);
-		if (op->flags & BCH_WRITE_submitted)
-			goto out_nofs_restore;
-	}
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
+
+	if (unlikely(op->opts.nocow &&
+		     c->opts.nocow_enabled) &&
+	    bch2_nocow_write(op))
+		return;
 again:
-	memset(&op->failed, 0, sizeof(op->failed));
+	op->wbio.failed.nr = 0;
 
 	do {
 		struct bkey_i *key_to_write;
@@ -1527,29 +1584,35 @@ again:
 					BKEY_EXTENT_U64s_MAX))
 			break;
 
-		/*
-		 * The copygc thread is now global, which means it's no longer
-		 * freeing up space on specific disks, which means that
-		 * allocations for specific disks may hang arbitrarily long:
-		 */
-		ret = bch2_trans_do(c,
-			bch2_alloc_sectors_start_trans(trans,
-				op->target,
-				op->opts.erasure_code && !(op->flags & BCH_WRITE_cached),
-				op->write_point,
-				&op->devs_have,
-				op->nr_replicas,
-				op->nr_replicas_required,
-				op->watermark,
-				op->flags,
-				&op->cl, &wp));
-		if (unlikely(ret)) {
-			if (bch2_err_matches(ret, BCH_ERR_operation_blocked))
+		CLASS(btree_trans, trans)(c);
+		struct alloc_request *req;
+		ret = lockrestart_do(trans, ({
+			req = alloc_request_get(trans,
+						op->target,
+						op->opts.erasure_code && !(op->flags & BCH_WRITE_cached),
+						&op->devs_have,
+						op->nr_replicas,
+						op->opts.data_replicas,
+						op->watermark,
+						op->flags,
+						&op->cl);
+			PTR_ERR_OR_ZERO(req) ?:
+			bch2_alloc_sectors_req(trans, req, op->write_point, &wp);
+		}));
+		bch2_trans_unlock_long(trans);
+		if (bch2_err_matches(ret, BCH_ERR_operation_blocked)) {
+			if (!wait_on_allocator_sync)
 				break;
 
-			goto err;
+			bch2_wait_on_allocator(c, req, ret, &op->cl);
+			__bch2_write_index(op);
+			op->wbio.failed.nr = 0;
+			continue;
 		}
 
+		if (unlikely(ret))
+			goto err;
+
 		EBUG_ON(!wp);
 
 		bch2_open_bucket_get(c, wp, &op->open_buckets);
@@ -1561,10 +1624,12 @@ err:
 			op->flags |= BCH_WRITE_submitted;
 
 			if (unlikely(ret < 0)) {
-				if (!(op->flags & BCH_WRITE_alloc_nowait))
-					bch2_write_op_error(op, op->pos.offset,
-							    "%s(): %s", __func__, bch2_err_str(ret));
 				op->error = ret;
+
+				/* Extra info on errors from the allocator: */
+				if (!(op->flags & BCH_WRITE_move))
+					bch2_write_op_error(op, true, op->pos.offset,
+							    "%s(): %s", __func__, bch2_err_str(ret));
 				break;
 			}
 		}
@@ -1582,17 +1647,8 @@ err:
 					  key_to_write, false);
 	} while (ret);
 
-	/*
-	 * Sync or no?
-	 *
-	 * If we're running asynchronously, wne may still want to block
-	 * synchronously here if we weren't able to submit all of the IO at
-	 * once, as that signals backpressure to the caller.
-	 */
-	if ((op->flags & BCH_WRITE_sync) ||
-	    (!(op->flags & BCH_WRITE_submitted) &&
-	     !(op->flags & BCH_WRITE_in_worker))) {
-		bch2_wait_on_allocator(c, &op->cl);
+	if (op->flags & BCH_WRITE_sync) {
+		closure_sync(&op->cl);
 
 		__bch2_write_index(op);
 
@@ -1603,8 +1659,6 @@ err:
 		bch2_write_queue(op, wp);
 		continue_at(&op->cl, bch2_write_index, NULL);
 	}
-out_nofs_restore:
-	memalloc_nofs_restore(nofs_flags);
 }
 
 static void bch2_write_data_inline(struct bch_write_op *op, unsigned data_len)
@@ -1615,7 +1669,7 @@ static void bch2_write_data_inline(struct bch_write_op *op, unsigned data_len)
 	unsigned sectors;
 	int ret;
 
-	memset(&op->failed, 0, sizeof(op->failed));
+	op->wbio.failed.nr = 0;
 
 	op->flags |= BCH_WRITE_wrote_data_inline;
 	op->flags |= BCH_WRITE_submitted;
@@ -1676,11 +1730,12 @@ CLOSURE_CALLBACK(bch2_write)
 	struct bch_fs *c = op->c;
 	unsigned data_len;
 
-	if (trace_io_write_enabled()) {
-		CLASS(printbuf, buf)();
-		bch2_write_op_to_text(&buf, op);
-		trace_io_write(c, buf.buf);
-	}
+	if (!(op->flags & BCH_WRITE_move))
+		event_add_trace(c, data_write, bio_sectors(bio), buf,
+				bch2_write_op_to_text(&buf, op));
+	else
+		event_add_trace(c, data_update_write, bio_sectors(bio), buf,
+				bch2_write_op_to_text(&buf, op));
 
 	EBUG_ON(op->cl.parent);
 	BUG_ON(!op->nr_replicas);
@@ -1692,13 +1747,12 @@ CLOSURE_CALLBACK(bch2_write)
 	if (op->flags & BCH_WRITE_only_specified_devs)
 		op->flags |= BCH_WRITE_alloc_nowait;
 
-	op->nr_replicas_required = min_t(unsigned, op->nr_replicas_required, op->nr_replicas);
 	op->start_time = local_clock();
 	bch2_keylist_init(&op->insert_keys, op->inline_keys);
 	wbio_init(bio)->put_bio = false;
 
 	if (unlikely(bio->bi_iter.bi_size & (c->opts.block_size - 1))) {
-		bch2_write_op_error(op, op->pos.offset, "misaligned write");
+		bch2_write_op_error(op, false, op->pos.offset, "misaligned write");
 		op->error = bch_err_throw(c, data_write_misaligned);
 		__WARN();
 		goto err;
@@ -1715,8 +1769,6 @@ CLOSURE_CALLBACK(bch2_write)
 		goto err;
 	}
 
-	if (!(op->flags & BCH_WRITE_move))
-		this_cpu_add(c->counters[BCH_COUNTER_io_write], bio_sectors(bio));
 	bch2_increment_clock(c, bio_sectors(bio), WRITE);
 
 	data_len = min_t(u64, bio->bi_iter.bi_size,
@@ -1739,14 +1791,14 @@ err:
 		op->end_io(op);
 }
 
-static const char * const bch2_write_flags[] = {
+const char * const bch2_write_flags[] = {
 #define x(f)	#f,
 	BCH_WRITE_FLAGS()
 #undef x
 	NULL
 };
 
-void bch2_write_op_to_text(struct printbuf *out, struct bch_write_op *op)
+void __bch2_write_op_to_text(struct printbuf *out, struct bch_write_op *op)
 {
 	if (!out->nr_tabstops)
 		printbuf_tabstop_push(out, 32);
@@ -1764,10 +1816,11 @@ void bch2_write_op_to_text(struct printbuf *out, struct bch_write_op *op)
 	prt_bitflags(out, bch2_write_flags, op->flags);
 	prt_newline(out);
 
+	prt_printf(out, "watermark:\t%s\n", bch2_watermarks[op->watermark]);
+
 	prt_printf(out, "nr_replicas:\t%u\n", op->nr_replicas);
-	prt_printf(out, "nr_replicas_required:\t%u\n", op->nr_replicas_required);
 	prt_printf(out, "devs_have:\t");
-	bch2_devs_list_to_text(out, &op->devs_have);
+	bch2_devs_list_to_text(out, op->c, &op->devs_have);
 	prt_newline(out);
 
 	prt_printf(out, "opts:\t");
@@ -1776,16 +1829,23 @@ void bch2_write_op_to_text(struct printbuf *out, struct bch_write_op *op)
 
 	prt_printf(out, "ref:\t%u\n", closure_nr_remaining(&op->cl));
 	prt_printf(out, "ret\t%s\n", bch2_err_str(op->error));
+}
+
+void bch2_write_op_to_text(struct printbuf *out, struct bch_write_op *op)
+{
+	__bch2_write_op_to_text(out, op);
 
 	if (op->flags & BCH_WRITE_move) {
+		guard(printbuf_indent)(out);
 		prt_printf(out, "update:\n");
 		guard(printbuf_indent)(out);
 		struct data_update *u = container_of(op, struct data_update, op);
 		bch2_data_update_opts_to_text(out, u->op.c, &u->op.opts, &u->opts);
 		prt_newline(out);
 
-		prt_str_indented(out, "old key:\t");
+		prt_str(out, "old key:\t");
 		bch2_bkey_val_to_text(out, u->op.c, bkey_i_to_s_c(u->k.k));
+		prt_newline(out);
 	}
 }
 
