@@ -11,6 +11,7 @@
 
 #include "alloc/backpointers.h"
 #include "alloc/buckets_waiting_for_journal.h"
+#include "alloc/discard.h"
 #include "alloc/disk_groups.h"
 #include "alloc/foreground.h"
 #include "alloc/replicas.h"
@@ -27,11 +28,12 @@
 #include "data/checksum.h"
 #include "data/compress.h"
 #include "data/copygc.h"
-#include "data/ec.h"
+#include "data/ec/create.h"
+#include "data/ec/init.h"
 #include "data/move.h"
 #include "data/nocow_locking.h"
 #include "data/read.h"
-#include "data/reconcile.h"
+#include "data/reconcile/work.h"
 #include "data/write.h"
 
 #ifdef CONFIG_DEBUG_FS
@@ -434,6 +436,7 @@ void bch2_fs_read_only(struct bch_fs *c)
 		bch2_verify_accounting_clean(c);
 	} else {
 		/* Make sure error counts/counters are persisted */
+		guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 		guard(mutex)(&c->sb_lock);
 		bch2_write_super(c);
 
@@ -613,6 +616,7 @@ static void __bch2_fs_free(struct bch_fs *c)
 	bch2_fs_errors_exit(c);
 	bch2_fs_encryption_exit(c);
 	bch2_fs_ec_exit(c);
+	bch2_fs_data_update_exit(c);
 	bch2_fs_counters_exit(c);
 	bch2_fs_copygc_exit(c);
 	bch2_fs_compress_exit(c);
@@ -947,7 +951,8 @@ static int bch2_fs_opt_version_init(struct bch_fs *c, struct printbuf *out)
 	if (c->opts.journal_rewind)
 		prt_printf(out, "rewinding journal, fsck required\n");
 
-	scoped_guard(mutex, &c->sb_lock) {
+	scoped_guard(memalloc_flags, PF_MEMALLOC_NOFS) {
+		guard(mutex)(&c->sb_lock);
 		struct bch_sb_field_ext *ext = bch2_sb_field_get(c->disk_sb.sb, ext);
 
 		__le64 now = cpu_to_le64(ktime_get_real_seconds());
@@ -998,6 +1003,18 @@ static int bch2_fs_opt_version_init(struct bch_fs *c, struct printbuf *out)
 		if (c->sb.version_upgrade_complete < bcachefs_metadata_version_autofix_errors)
 			SET_BCH_SB_ERROR_ACTION(c->disk_sb.sb, BCH_ON_ERROR_fix_safe);
 
+		unsigned extent_bp_shift_needed = ilog2(c->opts.encoded_extent_max >> 9) + 1;
+		if (extent_bp_shift_needed > c->sb.extent_bp_shift) {
+			prt_printf(out, "extent_bp_shift too small: must repair backpointers\n");
+			SET_BCH_SB_EXTENT_BP_SHIFT(c->disk_sb.sb, extent_bp_shift_needed);
+			ext->recovery_passes_required[0] |=
+				cpu_to_le64(bch2_recovery_passes_to_stable(BIT_ULL(BCH_RECOVERY_PASS_check_extents_to_backpointers)));
+			ext->recovery_passes_required[0] |=
+				cpu_to_le64(bch2_recovery_passes_to_stable(BIT_ULL(BCH_RECOVERY_PASS_check_backpointers_to_extents)));
+			__set_bit_le64(BCH_FSCK_ERR_backpointer_to_missing_ptr, ext->errors_silent);
+			__set_bit_le64(BCH_FSCK_ERR_ptr_to_missing_backpointer, ext->errors_silent);
+		}
+
 		/* Don't write the superblock, defer that until we go rw */
 	}
 
@@ -1061,6 +1078,7 @@ static int bch2_fs_init(struct bch_fs *c, struct bch_sb *sb,
 
 	refcount_set(&c->ro_ref, 1);
 	init_waitqueue_head(&c->ro_ref_wait);
+	mutex_init(&c->opt_change_lock);
 
 	for (unsigned i = 0; i < BCH_TIME_STAT_NR; i++)
 		bch2_time_stats_init(&c->times[i]);
@@ -1095,8 +1113,10 @@ static int bch2_fs_init(struct bch_fs *c, struct bch_sb *sb,
 
 	try(bch2_fs_capacity_init(c));
 
-	scoped_guard(mutex, &c->sb_lock)
+	scoped_guard(memalloc_flags, PF_MEMALLOC_NOFS) {
+		guard(mutex)(&c->sb_lock);
 		try(bch2_sb_to_fs(c, sb));
+	}
 
 	/* Compat: */
 	if (le16_to_cpu(sb->version) <= bcachefs_metadata_version_inode_v2 &&
@@ -1148,11 +1168,11 @@ static int bch2_fs_init(struct bch_fs *c, struct bch_sb *sb,
 	try(bch2_fs_buckets_waiting_for_journal_init(c));
 	try(bch2_fs_compress_init(c));
 	try(bch2_fs_counters_init(c));
+	try(bch2_fs_data_update_init(c));
 	try(bch2_fs_ec_init(c));
 	try(bch2_fs_errors_init(c));
 	try(bch2_fs_encryption_init(c));
 	try(bch2_fs_io_read_init(c));
-	try(bch2_fs_reconcile_init(c));
 	try(bch2_fs_vfs_init(c));
 	try(bch2_io_clock_init(&c->io_clock[READ]));
 	try(bch2_io_clock_init(&c->io_clock[WRITE]));
@@ -1189,7 +1209,8 @@ static int bch2_fs_init(struct bch_fs *c, struct bch_sb *sb,
 			&c->clock_journal_res,
 			(sizeof(struct jset_entry_clock) / sizeof(u64)) * 2);
 
-	scoped_guard(mutex, &c->sb_lock) {
+	scoped_guard(memalloc_flags, PF_MEMALLOC_NOFS) {
+		guard(mutex)(&c->sb_lock);
 		if (!bch2_sb_field_get_minsize(&c->disk_sb, ext,
 				sizeof(struct bch_sb_field_ext) / sizeof(u64)))
 			return bch_err_throw(c, ENOSPC_sb);
@@ -1209,13 +1230,6 @@ static int bch2_fs_init(struct bch_fs *c, struct bch_sb *sb,
 		try(bch2_fs_opt_version_init(c, out));
 	}
 
-	/*
-	 * just make sure this is always allocated if we might need it - mount
-	 * failing due to kthread_create() failing is _very_ annoying
-	 */
-	if (go_rw_in_recovery(c))
-		try(bch2_fs_init_rw(c));
-
 	scoped_guard(mutex, &bch2_fs_list_lock)
 		try(bch2_fs_online(c));
 
@@ -1239,6 +1253,17 @@ static struct bch_fs *bch2_fs_alloc(struct bch_sb *sb, struct bch_opts *opts,
 	return c;
 }
 
+void bch2_missing_devs_to_text(struct printbuf *out, struct bch_fs *c)
+{
+	prt_printf(out, "Missing devices\n");
+	for_each_member_device(c, ca)
+		if (!bch2_dev_is_online(ca) && bch2_dev_has_data(c, ca)) {
+			prt_printf(out, "Device %u\n", ca->dev_idx);
+			guard(printbuf_indent)(out);
+			bch2_member_to_text_short(out, c, ca);
+		}
+}
+
 static int bch2_fs_may_start(struct bch_fs *c, struct printbuf *err)
 {
 	unsigned flags = 0;
@@ -1270,17 +1295,10 @@ static int bch2_fs_may_start(struct bch_fs *c, struct printbuf *err)
 	}
 	}
 
-	if (!bch2_can_read_fs_with_devs(c, c->devs_online, flags, err) ||
+	if (!bch2_can_read_fs_with_devs(c, &c->devs_online, flags, err) ||
 	    (!c->opts.read_only &&
 	     !bch2_can_write_fs_with_devs(c, c->allocator.rw_devs[0], flags, err))) {
-		prt_printf(err, "Missing devices\n");
-		for_each_member_device(c, ca)
-			if (!bch2_dev_is_online(ca) && bch2_dev_has_data(c, ca)) {
-				prt_printf(err, "Device %u\n", ca->dev_idx);
-				guard(printbuf_indent)(err);
-				bch2_member_to_text_short(err, c, ca);
-			}
-
+		bch2_missing_devs_to_text(err, c);
 		return bch_err_throw(c, insufficient_devices_to_start);
 	}
 
@@ -1302,6 +1320,16 @@ static int __bch2_fs_start(struct bch_fs *c, struct printbuf *err)
 
 	try(bch2_fs_may_start(c, err));
 
+	try(bch2_fs_reconcile_init(c));
+	try(bch2_fs_counters_init_late(c));
+
+	/*
+	 * just make sure this is always allocated if we might need it - mount
+	 * failing due to kthread_create() failing is _very_ annoying
+	 */
+	if (go_rw_in_recovery(c))
+		try(bch2_fs_init_rw(c));
+
 	/*
 	 * check mount options as early as possible; some can only be checked
 	 * after starting
@@ -1379,7 +1407,8 @@ int bch2_fs_resize_on_mount(struct bch_fs *c)
 				return ret;
 			}
 
-			scoped_guard(mutex, &c->sb_lock) {
+			scoped_guard(memalloc_flags, PF_MEMALLOC_NOFS) {
+				guard(mutex)(&c->sb_lock);
 				struct bch_member *m =
 					bch2_members_v2_get_mut(c->disk_sb.sb, ca->dev_idx);
 				m->nbuckets = cpu_to_le64(new_nbuckets);
